/*
#define PRDEBUG
*/

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#   define MODVERSIONS
#endif

#if defined(CONFIG_SMP) && !defined(__SMP__)
#   define __SMP__
#endif

#ifdef MODVERSIONS
#   include <linux/modversions.h>
#endif


#define MODULE
#define __KERNEL__ 

#define FSNAME "cdfs"
#define VERSION "V0.3d"

#define BANNER       FSNAME##"\t"##VERSION
#define EMPTYMESSAGE BANNER##"\n\tNo CD mounted\n\n"


#ifdef PRDEBUG
# define PRINT(format, arg...) printk(FSNAME "-> " format, ## arg)
#else
# define PRINT(format, arg...) 
#endif


#include <linux/types.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <linux/locks.h>
#include <linux/init.h>
#include <linux/cdrom.h>
#include <linux/iso_fs.h>
#include <linux/time.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <asm/uaccess.h>

#define SECTORSIZE 2048
#define SECTORBITS 11

#define CDFS_MAGIC 0xCDDA
#define CDFS_MAXFN 128

#define TRUE 1
#define FALSE 0

#define MAX_TRACKS 99
#define WAV_HEADER_SIZE 44

#define READ_AHEAD 5

#define NONE  0
#define AUDIO 1
#define DATA  2
#define BOOT  3
#define HFS   4

#define AUDIO_NAME       "track-%02d.wav"
#define DATA_NAME_ISO    "sessions_1-%d.iso"
#define DATA_NAME_SINGLE "session-%d.iso"
#define DATA_NAME_UNK    "data-%d.unknown"

#define UID   500
#define GID   500
#define MODE  (S_IFREG | S_IRUSR | S_IRGRP | S_IROTH) 

typedef struct _iso_track_ {
  unsigned start;
  unsigned stop;
} iso_track;

typedef struct _track_info {
  unsigned size;                              /* in bytes */
  unsigned start_lba;
  unsigned stop_lba;
  unsigned hfs_offset;                        /* in 512 bytes */
  unsigned track_size;                        /* in bytes */
  unsigned iso_size;                          /* in bytes */
  unsigned type;                              /* audio, data of boot */
  char name[99];
  unsigned inode;
  time_t time;                                /* only for data tracks */
  struct iso_primary_descriptor * iso_info;   /* only for data tracks */
  char bootID[99];
} track_info;

typedef struct _cd_ {
  mode_t mode;
  gid_t gid;
  uid_t uid;    
  int single;
  unsigned size;                      /* bytes */
  unsigned tracks;                    /* aantal tracks */
  track_info track[MAX_TRACKS+1+1];   /* info per track + leadout + bootimage*/
  struct inode inode_fake;
  int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); 
  int nr_iso_sessions;
  iso_track lba_iso_sessions[MAX_TRACKS+1];
  char * audio_cache;
  int sector_cache;                   /* address of first sector in cache */
} cd;

typedef unsigned char byte; 

int read_proc(char *buf, char **start, off_t offset, int len, int *eof, void *data );
  
struct inode_operations cdda_file_inode_operations;
struct inode_operations cdfs_file_inode_operations;
struct inode_operations cdXA_file_inode_operations;
struct inode_operations cdhfs_file_inode_operations;

time_t constructtime(char * time);
unsigned constructsize(char * size);
void constructMSFsize(char * result, unsigned length);
int cdfs_ioctl(struct super_block *s, int cmd, unsigned long arg);
struct iso_primary_descriptor * get_iso_info(struct super_block *sb, int track_no);
int get_hfs_info(struct super_block *sb, unsigned track);
void check_bootable(struct super_block *sb);
